/*
  Author: Greg (Botzilla)
  http://botzilla.free.fr/

  Date: 2002/11/11

  This is a simualtion program for PID regulation of a small robot.
  There is no local variables, and only short or char operations
  because it will be adapt for the real implementation on a PIC 16F877

  Compilation: gcc pid.c -o pid -lm
*/

//==============================================================
//==============================================================
/*
  2002/11/11:
  Param moteur
  Gauche: 1. / 20.
  Droit: 10. / 20.

  Avec P (Kp=100)
  oscil et biais

  Avec P et D (Kp=100 / Kd=80)
  convergence bonne mais avec un biais (4cm)

  Avec P, I et D (Kp=100 / Ki=4 / Kd=80)
  convergence bonne avec un tres petit biais (1mm)
*/
//==============================================================
//==============================================================

#define __P__ 
#define __I__ 
//#define __D__

//PID : Cr=Cr+(P/_Kp+I/_Ki+D/Kd) and Cl=Cl-(P/_Kp+I/_Ki+D/Kd) 
unsigned char _Kp=100;
unsigned char _Ki=4;
unsigned char _Kd=80;

//Motor simulation
double _Mrt=10.;
double _Mrc=20.;
double _Mlt=1.;
double _Mlc=20.;

//==============================================================
//==============================================================

#include <stdlib.h>
#include <stdio.h>
#include <math.h>

//General constant
unsigned short _Neutre=3000;

//Robot constant
unsigned short _D=3;//3 mm
unsigned short _R=300;//mm

// initial position
unsigned short _X0=3000;//mm
unsigned short _Y0=3000;//mm

// position
unsigned short _X;//mm
unsigned short _Y;//mm
unsigned short _A;//10-2 rd
unsigned char _Sa;
unsigned char _Sap;

//Velocity
unsigned short _Nr;//Nb Right
unsigned short _Nl;//Nb Left
unsigned short _V;//Average velocity
unsigned short _Va;//Angle velocity
unsigned char _SVa;
unsigned char _SVap;
unsigned short _Vx;
unsigned char _Sx;
unsigned short _Vy;
unsigned short _Vyp;
unsigned short _Ay;
unsigned char _SAy;

unsigned char _Cr;
unsigned char _Cl;

unsigned short tempP;
unsigned short tempI;
unsigned short tempD;
unsigned short temp;
unsigned short tempPID;
unsigned short NeutrePID;
unsigned short tempVx;
unsigned short tempVy;

void init()
{
  _X=_X0;//mm
  _Y=_Y0;//mm
  _A=0;//10-2 rd
  _Sa=1;//0: neg / 1: pos
  _Cr=50;//% PWM
  _Cl=50;//% PWM
  _Vy=0;//mm
}

// return 100+100*cos(a/100)
unsigned char _cos(unsigned short a)
{
  double c=cos((double)a/100.)*100.;
  return (unsigned char)(100.+c);
}

//return 100*sin(a/100)
unsigned char _sin(unsigned short a)
{
  double s=abs(sin((double)a/100.)*100.);
  return (unsigned char)s;
}

//simulation of the motors reactions
void run(unsigned char cr, unsigned char cl,
	 unsigned short *nr,unsigned short *nl)
{
  double vr=(exp(_Mrt)-exp(-_Mrt*(cr-100.)/100.))/(exp(_Mrt)-exp(0.))*_Mrc;
  double vl=(exp(_Mlt)-exp(-_Mlt*(cl-100.)/100.))/(exp(_Mlt)-exp(0.))*_Mlc;
  (*nr)=(unsigned short)vr;
  (*nl)=(unsigned short)vl;
  printf("_Cr=%d _Cl=%d _Nr=%d _Nl=%d ",cr,cl,(*nr),(*nl));
}

//estimation of X,Y,A
void update_position(unsigned char verbose)
{
  //backup the previous results
  _Sap=_Sa;
  _SVap=_SVa;
  _Vyp=_Vy;
  //velocity estimation
  _V=_Nr+_Nl;//Nb/T
  _V=_V*_D;//mm/T
  _V=_V/2;//Average velocity (mm/period)
  if (verbose) printf("_V=%d ",_V);
  if (_Nl>_Nr)
    {
      _Va=_Nl-_Nr;
      _SVa=1;
    }
  else
    {
      _Va=_Nr-_Nl;
      _SVa=0;
    }
  _Va=_Va*_D;
  _Va=_Va*100;
  _Va=_Va/_R;//Angle velocity (10-2 rd)
  if (verbose) printf("_Va=%d _SVa=%d ",_Va,_SVa);
  //Angle estimation
  if (_SVa==_Sa)
    {
      _A=_A+_Va;
    }
  else
    {
      if (_A>_Va)
	{
	  _A=_A-_Va;
	}
      else
	{
	  _Sa=_SVa;
	  _A=_Va-_A;
	}
    }
  //End of the angle estimation (10-2 rd)
  if (verbose) printf("_A=%d _Sa=%d ",_A,_Sa);
  //deplacement estimation (10-2 mm)
  _Vx=_cos(_A);
  if (_Vx<100)//Vx<0
    {
      _Sx=0;
      _Vx=100-_Vx;
    }
  else
    {
      _Sx=1;
      _Vx=_Vx-100;
    }
  _Vx=_Vx*_V;
  _Vy=_V*_sin(_A);
  if (verbose) printf("_Vx=%d _Vy=%d ",_Vx,_Vy);
  //new position estimation
  tempVx=_Vx/100;//10-2 mm -> mm
  tempVy=_Vy/100;//10-2 mm -> mm
  if (_Sx)//Vx>0
    {
      _X=_X+tempVx;
    }
  else
    {
      _X=_X-tempVx;
    }
  if (_Sa)//Vy>0
    {
      _Y=_Y+tempVy;
    }
  else
    {
      _Y=_Y-tempVy;
    }
  //end of the new position estimation (mm)
  if (verbose) printf("_X=%d _Y=%d ",_X,_Y);
  if (verbose) printf("_Sap=%d _Sa=%d ",_SVap,_SVa);
  //lateral acceleration estimation
  if (_Sap==_Sa)
    {
      _SAy=_SVa;
      if (_Vy>_Vyp)
	{
	  _Ay=_Vy-_Vyp;
	}
      else
	{
	  _Ay=_Vyp-_Vy;
	}
     
    }
  else
    {
      _Ay=_Vy+_Vyp;
      if (_Vy>_Vyp)
	{
	  _SAy=_SVa;
	}
      else
	{
	  _SAy=_SVap;
	}
    }
  if (verbose) printf("_Ay=%d _SAy=%d ",_Ay,_SAy);
}

// PID regulation update
void update_command()
{
  NeutrePID=0;
  //P factor : lateral velocity
#ifdef __P__
  tempP=_Neutre;
  if (_Sa)
    {
      tempP=tempP+_Vy;
    }
  else 
    {
      tempP=tempP-_Vy;
    }
  tempP=tempP/_Kp;
  NeutrePID=_Neutre/_Kp;
#else
  tempP=0;
#endif
  //I factor : lateral error
#ifdef __I__
  tempI=_Y;
  tempI=tempI/_Ki;
  NeutrePID=NeutrePID+_Y0/_Ki;
#else
  tempI=0;
#endif
  //D fector : lateral acceleration
#ifdef __D__
  tempD=_Neutre;
  if (_SAy)
    {
      tempD=tempD+_Ay;
    }
  else
    {
      tempD=tempD-_Ay;
    }
  tempD=tempD/_Kd;
  NeutrePID=NeutrePID+_Neutre/_Kd;
#else
  tempD=0;
#endif
  //End of the 3 PID factors estimation (with their ponderations)
  printf("tempP=%d tempI=%d tempD=%d ",tempP,tempI,tempD-_Neutre/_Kd);
  tempPID=tempP+tempI+tempD;

  if (NeutrePID>tempPID)
    {
      //PID<0 -> increase Cl / decrease Cr
      tempPID=NeutrePID-tempPID;
      printf("tempPID=%d ",tempPID);
      if (_Cr<tempPID)
	{
	  _Cr=0;
	}
      else
	{
	  _Cr=_Cr-tempPID;
	}
      tempPID=tempPID+_Cl;
      if (tempPID>100)
	{
	  _Cl=100;
	}
      else
	{
	  _Cl=tempPID;
	}
    }
  else
    {
      //PID>0 -> increase Cr / decrease Cl
      tempPID=tempPID-NeutrePID;
      printf("tempPID=%d ",tempPID);
      if (_Cl<tempPID)
	{
	  _Cl=0;
	}
      else
	{
	  _Cl=_Cl-tempPID;
	}
      tempPID=tempPID+_Cr;
      if (tempPID>100)
	{
	  _Cr=100;
	}
      else
	{
	  _Cr=tempPID;
	}
    }
  //End of Cr / Cl estimation
  printf("_Cr=%d _Cl=%d ",_Cr,_Cl);
}

int main(void)
{
  int i;
  printf("i=Date _Cr=Cr _Cl=Cl _Nr=Nr _Nl=Nl _V=V _Va=Va _Sva=SVa _A=A _Sa=Sa _Vx=Vx _Vy=Vy _X=X _Y=Y _Sap=Sap _Sa=Sa _Ay=Ay _SAy=SAy tempP=tempP tempI=tempI tempPID=tempPID _Cr=Cr _Cl=Cl\n");
  init();
  for(i=0;i<200;++i)
    {
      printf("i=%d ",i);
      run(_Cr,_Cl,&_Nr,&_Nl);
      update_position(1);
      update_command();
      printf("\n");
    }
  return 0;
}
